Attribute VB_Name = "basDPlay"
'DPlay, Copyright 1998, Patrice Scribe
'Modified: 12-10-98 Dan Jones
'  - Added parsing of system messages when players leave and join
'  - Added conditional compilation for using Unicode or Ansi strings
'  - Added GUIDs for DirectPlay4 and DirectPlay4A
'  - Tested with TCP/IP and IPX/SPX connections
'
'A simple DirectPlay4 sample
'Needs DirectX support v 0.6 or later, Win32 support v 0.2 or later
'Tested with serial connection only

Option Explicit

' Not declared properly in the Win32 type library ?
Private Declare Function lstrcpy Lib "kernel32" (a As Any, b As Any) As Long

Private Const DPSYS_CREATEPLAYERORGROUP = 3
Private Const DPSYS_DESTROYPLAYERORGROUP = 5
Private Const DPSYS_ADDPLAYERTOGROUP = 7
Private Const DPSYS_DELETEPLAYERFROMGROUP = 33
Private Const DPSYS_SESSIONLOST = 49

Public dp As directx.IDirectPlay3           ' Main DirectPlay object
Public ChoosenConnection As Long    ' Choosen connection

Dim CLSID_DirectPlay As Win32.GUID  ' DirectPlay class ID
'Dim IID_IDirectPlay3A As Win32.Guid  ' DirectPlay3A interface ID

#If UseUNICode = -1 Then
Dim IID_IDirectPlay4 As Win32.GUID
#Else
Dim IID_IDirectPlay4A As Win32.GUID  ' DirectPlay3A interface ID
#End If

Dim AppId As Win32.GUID             ' This app ID (created by uuidgen)

' Enumerated connections
Public Type TConnection
    Connection() As Byte    ' Connection data
    Name As String          ' Connection name
End Type
Public Type Buffer
    Data() As Byte
End Type

Public Connections(0 To 9) As TConnection
Public ConnectionCount As Long

' Enumerated sessions
Public Type TSession
    Description As DPSESSIONDESC2
    Name As String
    Password As String
End Type
Public Sessions(1 To 10) As TSession
Public SessionCount As Long

Sub Main()
    Dim ShortName As String     ' Player short name
    Dim LongName As String      ' Player long name
    Dim PlayerName As DPNAME    ' Player name structure
    Dim hEvent As Long          ' Event raised when incoming data
    Dim PlayerId As Long        ' Player id
    Dim ReceivedData() As Byte
    Dim DataSize As Long        ' Data length
    
    Dim MessageCount As Long    ' Number of messages
    Dim sd As DPSESSIONDESC2
    Dim SessionName As String
    Dim From As DPID
    Dim CreatePlayerOrGroup As DPMSG_CREATEPLAYERORGROUP
    Dim DestroyPlayerOrGroup As DPMSG_DESTROYPLAYERORGROUP
    Dim Data As String
    Dim i As Long
    
    ' Initialize GUID values
    IIDFromString StrPtr("{D1EB6D20-8923-11D0-9D97-00A0C90A43CB}"), CLSID_DirectPlay
    'IIDFromString StrPtr("{133EFE41-32DC-11D0-9CFB-00A0C90A43CB}"), IID_IDirectPlay3A
    
    'Regardless of whether we're using UNICode or not, the IIDFromString function
    'wants a UNICode string. That's why we won't put a StrConv around the string to
    'convert it to ANSI
    #If UseUNICode = -1 Then
    IIDFromString StrPtr("{0AB1C530-4745-11D1-A7A1-0000F803ABFC}"), IID_IDirectPlay4
    #Else
    IIDFromString StrPtr("{0AB1C531-4745-11D1-A7A1-0000F803ABFC}"), IID_IDirectPlay4A
    #End If
    
    IIDFromString StrPtr("{17553DA0-02EE-11D2-A098-9C9501C1441E}"), AppId
    
    #If UseUNICode = -1 Then
    'Create the DirectPlay object (DirectPlayCreate is said to be obsolete)
    CoCreateInstance CLSID_DirectPlay, 0, 1, IID_IDirectPlay4, dp
    #Else
    'Create the DirectPlay object (DirectPlayCreate is said to be obsolete)
    CoCreateInstance CLSID_DirectPlay, 0, 1, IID_IDirectPlay4A, dp
    #End If
    
    ' Enumerate connections
    dp.EnumConnections ByVal 0&, AddressOf EnumConnectionsCallback, ByVal 0&, 0
    
    frmDPlay.Show vbModal
    If ChoosenConnection = -1 Then
        Set dp = Nothing
        End
    End If
    
    ' Initialize choosen connection
    dp.InitializeConnection Connections(ChoosenConnection).Connection(0), 0
    
    ' Enumerate sessions
    frmDPlayMsg.Show
    frmDPlayMsg.Message "Enumerating..."
    sd.dwSize = Len(sd)
    sd.guidApplication = AppId
    dp.EnumSessions sd, 0, AddressOf EnumSessionsCallback, ByVal 0&, DPENUMSESSIONS_AVAILABLE
    
    ' If no session found
    If SessionCount = 0 Then
        ' Create a session
        SessionName = "VBA51"
        With sd
            .dwFlags = DPSESSION_KEEPALIVE
            .dwSize = Len(sd)
            #If UseUNICode = -1 Then
            .lpszSessionName = StrPtr(SessionName)
            #Else
            .lpszSessionName = StrPtr(StrConv(SessionName, vbFromUnicode))
            #End If
        End With
        dp.Open sd, DPOPEN_CREATE
        frmDPlayMsg.Message "Session created..."
    Else
        ' Join a session
        sd = Sessions(1).Description
        
        #If UseUNICode = -1 Then
        sd.lpszSessionName = StrPtr(Sessions(1).Name)
        #Else
        sd.lpszSessionName = StrPtr(StrConv(Sessions(1).Name, vbFromUnicode))
        #End If
        
        dp.Open sd, DPOPEN_JOIN
        Msg "Joined " & Sessions(1).Name
    End If
    
    ' Initialize player data
    ShortName = "Pat"
    LongName = "Patrice Scribe"
    PlayerName.dwSize = Len(PlayerName)
    
    #If UseUNICode = -1 Then
    PlayerName.lpszShortName = StrPtr(ShortName)
    PlayerName.lpszLongName = StrPtr(LongName)
    #Else
    PlayerName.lpszShortName = StrPtr(StrConv(ShortName, vbFromUnicode))
    PlayerName.lpszLongName = StrPtr(StrConv(LongName, vbFromUnicode))
    #End If
    
    ' Create an event for incoming data
    hEvent = CreateEvent(ByVal 0&, 0, 0, vbNullString)
    
    ' Create player
    dp.CreatePlayer PlayerId, PlayerName, hEvent, ByVal 0, 0, 0
    frmDPlayMsg.Show
    Do While DoEvents() <> 0
        Data = ""
        ' If event raised...
        If WaitForSingleObject(hEvent, 0) = 0 Then
            ' Read all messages
            dp.GetMessageCount PlayerId, MessageCount
            Do While MessageCount > 0
                From = DPID_ALLPLAYERS

                DataSize = 0

                On Error Resume Next
                Do
                    Err.Clear
                    ReDim ReceivedData(DataSize)
                    dp.Receive From, PlayerId, DPRECEIVE_ALL, ReceivedData(0), DataSize
                Loop Until Err.Number <> -2005467106
                MessageCount = MessageCount - 1

                If From = DPID_SYSMSG Then
                    Select Case ReceivedData(0)
                        Case DPSYS_CREATEPLAYERORGROUP
                            'Only copy as much data into the CreatePlayerOrGroup variable as it can hold
                            'If we use DataSize, we'll be putting in too much data
                            'Why? Because ReceivedData actually contains not only the structure, but the
                            'data that goes with the structure (in this case, the CreatePlayerOrGroup structure
                            'contains two pointers, one for each name. The string data for the two names also
                            'comes across as part of ReceivedData, tagged on the end)
                            '
                            'To verify this is what happens, uncomment the block of code following CopyMemory
                            CopyMemory CreatePlayerOrGroup, ReceivedData(0), Len(CreatePlayerOrGroup)
                            
                            'The following block of code will show that the bytes following the
                            'CreatePlayerOrGroup structure will contain the two strings pointed to by
                            'lpszShortName and lpszLongName
                            'Dim dummy() As Byte
                            'ReDim dummy(DataSize - Len(CreatePlayerOrGroup))
                            'CopyMemory dummy(0), ReceivedData(Len(CreatePlayerOrGroup)), DataSize - Len(CreatePlayerOrGroup)
                            'For i = 0 To DataSize - Len(CreatePlayerOrGroup) - 1
                            '    Debug.Print Chr$(dummy(i));
                            'Next i
                            
                            #If UseUNICode = -1 Then
                            CopyUniCode Data, CreatePlayerOrGroup.dpnName.lpszShortName
                            #Else
                            'We're going to add one to the number of spaces
                            Data = Space$(lstrlen(ByVal CreatePlayerOrGroup.dpnName.lpszShortName))
                            lstrcpy ByVal Data, ByVal CreatePlayerOrGroup.dpnName.lpszShortName
                            #End If
                            Msg Data & " has joined"
                            Data = ""
                        Case DPSYS_DESTROYPLAYERORGROUP
                            'Only copy as much data into the DestroyPlayerOrGroup variable as it can hold
                            'If we use DataSize, we'll be putting in too much data
                            'Why? Because ReceivedData actually contains not only the structure, but the
                            'data that goes with the structure (in this case, the CreatePlayerOrGroup structure
                            'contains two pointers, one for each name. The string data for the two names also
                            'comes across as part of ReceivedData, tagged on the end)
                            '
                            'To verify this is what happens, uncomment the block of code following CopyMemory
                            CopyMemory DestroyPlayerOrGroup, ReceivedData(0), Len(DestroyPlayerOrGroup)
                            
                            'The following block of code will show that the bytes following the
                            'CreatePlayerOrGroup structure will contain the two strings pointed to by
                            'lpszShortName and lpszLongName
                            'Dim dummy() As Byte
                            'ReDim dummy(DataSize - Len(DestroyPlayerOrGroup))
                            'CopyMemory dummy(0), ReceivedData(Len(DestroyPlayerOrGroup)), DataSize - Len(DestroyPlayerOrGroup)
                            'For i = 0 To DataSize - Len(DestroyPlayerOrGroup) - 1
                            '    Debug.Print Chr$(dummy(i));
                            'Next i
                            
                            #If UseUNICode = -1 Then
                            CopyUniCode Data, DestroyPlayerOrGroup.dpnName.lpszShortName
                            #Else
                            Data = Space$(lstrlen(ByVal DestroyPlayerOrGroup.dpnName.lpszShortName))
                            lstrcpy ByVal Data, ByVal DestroyPlayerOrGroup.dpnName.lpszShortName
                            #End If
                            Msg Data & " has left"
                            Data = ""
                        Case Else
                    End Select
                Else
                    'We'll divide the DataSize by 2 since DataSize represents the number
                    'of bytes. Since we passed in our string as a UNICode string - where
                    'each character is two bytes, we really only need to allocate half
                    'as many spaces
                    Data = Space$(DataSize / 2)
                    
                    CopyMemory ByVal Data, ReceivedData(0), DataSize
                    Msg "Received : " & Data
                    Data = ""
                End If
            Loop
        End If
        
        ' Send a random message when a key is pressed
        If GetAsyncKeyState(vbKeySpace) <> 0 Then
            Select Case Int(Rnd * 4 + 1)
                Case 1
                    Data = "Left" & Chr$(0)
                Case 2
                    Data = "Up" & Chr$(0)
                Case 3
                    Data = "Down" & Chr$(0)
                Case 4
                    Data = "Right" & Chr$(0)
                Case Else
                    Debug.Assert False
            End Select
        End If
        If Data <> "" Then
            'We'll send the string Data as a UNICode string. We don't really
            'have to worry about this too much except for when we give the
            'number of bytes - in this case, we have to multiply by two (since
            'each UNICode character takes up two bytes)
            dp.Send PlayerId, DPID_ALLPLAYERS, 0, ByVal Data, Len(Data) * 2
            
            Msg "Sent : " & Data
        End If
    Loop
    
    ' Clean up is probably to improve
    dp.DestroyPlayer PlayerId
    dp.Close
    Set dp = Nothing
End Sub


Public Function EnumSessionsCallback(ByVal lpSessionDescription As Long, TimeOut As Long, ByVal Flags As Long, ByVal lpContext As Long) As Long
Dim lSize As Long
Dim b(1) As Byte
Dim i As Long

    If lpSessionDescription <> 0 Then
        SessionCount = SessionCount + 1
        With Sessions(SessionCount)
            ' Copy connection data
            CopyMemory .Description, ByVal lpSessionDescription, Len(.Description)
            
            #If UseUNICode = -1 Then
            'Since we don't know how to copy a unicode string, we'll try it the hard way - two
            'bytes at a time until we hit a double-null (Chr$(0) & chr$(0))
            CopyUniCode .Name, .Description.lpszSessionName
            #Else
            'Although .Name will be a UNICode string, by using the Space$ function,
            'we'll get the right amount of memory by using lstrlen
            lSize = lstrlen(.Description.lpszSessionName)
            .Name = Space$(lSize)
            
            ' Copy name
            lstrcpy ByVal .Name, ByVal .Description.lpszSessionName
            #End If
            
            If .Description.lpszPassword <> 0 Then
                #If UseUNICode = -1 Then
                ' Copy password
                CopyUniCode .Password, .Description.lpszPassword
                #Else
                lSize = lstrlen(ByVal .Description.lpszPassword)
                .Password = Space$(lSize)
                ' Copy password
                lstrcpy ByVal .Password, ByVal .Description.lpszPassword
                #End If
            End If
        End With
    End If
    EnumSessionsCallback = 0    ' Stop enumeration
End Function

Function EnumConnectionsCallback(ByRef lpGuid As Win32.GUID, ByVal lpConnection As Long, ByVal dwConnectionSize As Long, ByVal lpName As Long, ByVal dwFlags As Long, ByVal lpContext As Long) As Long
Dim b(1000) As Byte
Dim i As Long
    Dim Name As DPNAME
    With Connections(ConnectionCount)
        ' Set connection data buffer size
        ReDim .Connection(dwConnectionSize)
        
        ' Copy connection data
        CopyMemory .Connection(0), ByVal lpConnection, dwConnectionSize
        
        ' Copy name structure
        CopyMemory Name, ByVal lpName, Len(Name)
        
        #If UseUNICode = -1 Then
        'Set name size
        CopyUniCode .Name, Name.lpszShortName
        #Else
        'Set name size
        .Name = Space$(lstrlen(Name.lpszShortName))
        
        'Copy name
        lstrcpy ByVal .Name, ByVal Name.lpszShortName
        #End If
        
    End With
    
    ConnectionCount = ConnectionCount + 1
    EnumConnectionsCallback = 1    ' Continue enumeration
End Function

Private Sub Msg(sString As String)
    frmDPlayMsg.Message sString
End Sub

Private Sub CopyUniCode(sString As String, lpszUniCodeString As Long)
'This sub will copy the unicode string pointed to by lpszUniCodeString into sString
'It does so by copying two bytes of data at a time, looking for a double-null
'which signifies the end of the string
Dim b(1) As Byte
Dim i As Long

    sString = ""
    i = 0
    CopyMemory b(0), ByVal lpszUniCodeString, 2
    Do While b(0) <> 0 Or b(1) <> 0
        sString = sString & Chr$(b(0))
        i = i + 2
        CopyMemory b(0), ByVal lpszUniCodeString + i, 2
    Loop
End Sub
